if [ -z ${CONDA_BUILD+x} ]; then
    source /Users/ktietz/demo/mc3/conda-bld/pymongo_1629276480182/work/build_env_setup.sh
fi
python setup.py install --single-version-externally-managed --record=record.txt